"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.engine.lighting import Light as __Light
from panda3d_gemstone.engine.fog import Fog as __Fog
from panda3d_gemstone.world.scene import Scene as __Scene
from panda3d_gemstone.framework import bootstrap as __bootstrap
from panda3d_gemstone.controllers.avatar import AvatarController as __AvatarController
from panda3d_gemstone import __version__ as package_version

__version__ = package_version.__version__

__singleton_list = [
    __bootstrap.create_singleton_entry('panda3d_gemstone.application.events.EventsManager'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.logging.handler.ApplicationLogger'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.framework.service.ServiceManager'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.framework.resource.ResourceManager'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.engine.collisions.CollisionSystem'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.engine.camera.CameraManager'),
    __bootstrap.create_singleton_entry('panda3d_gemstone.engine.audio.SoundManager'),
]

__class_list = [
    __bootstrap.create_class_entry('panda3d_gemstone.engine.lighting.Light'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.fog.Fog'),
    __bootstrap.create_class_entry('panda3d_gemstone.world.world.World'),
    __bootstrap.create_class_entry('panda3d_gemstone.world.streaming.StreamableWorld'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.skybox.Skybox'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.terrain.CPUTerrain'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.terrain.GPUTerrain'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.terrain.PaintedChannelTerrain'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.model.StaticModel'),
    __bootstrap.create_class_entry('panda3d_gemstone.world.streaming.StreamableStaticModel'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.model.AnimatedModel'), 
    __bootstrap.create_class_entry('panda3d_gemstone.engine.model.TextureModel'),
    __bootstrap.create_class_entry('panda3d_gemstone.world.entity.Entity'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.lighting.LightSet'),
    __bootstrap.create_class_entry('panda3d_gemstone.engine.text.Font'),
    __bootstrap.create_class_entry('panda3d_gemstone.game.avatar.Avatar'),
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.movement.MovementController'), 
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.movement.JumpController'), 
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.basic.FixedController'), 
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.camera.CameraCollisionController'), 
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.physics.PhysicsMovementController'),
    __bootstrap.create_class_entry('panda3d_gemstone.controllers.avatar.AvatarController')
]

__meta_list = [
    ('Light', 'Scene.insert', __Light.insert),
    ('Light', 'Scene.remove', __Light.remove),
    ('Light', 'editor_object', True),
    ('Fog', 'Scene.insert', __Fog.insert),
    ('Fog', 'Scene.remove', __Fog.remove),
    ('Fog', 'editor_object', True),
    ('Skybox', 'editor_object', True),
    ('CPUTerrain', 'editor_object', True),
    ('GPUTerrain', 'editor_object', True),
    ('StaticModel', 'editor_object', True),
    ('StreamableStaticModel', 'editor_object', True),
    ('AnimatedModel', 'editor_object', True),
    ('Font', 'Scene.insert', __Scene.insert_empty),
    #('JumpController', 'Controller.setup', __JumpController.set_ground_collision),
    #('CameraCollisionController', 'Controller.setup', __CameraCollisionController.set_target),
    ('AvatarController', 'Controller.setup', __AvatarController.setup)
]

# Perform setup operations on the gemstone module
__bootstrap.bootstrap_module(
    class_list = __class_list,
    meta_list=__meta_list,
    singleton_list=__singleton_list)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os

from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.engine import prc

__default_revision = prc.get_prc_string('gs-sc-default', '???')
__source_control_notify = get_notify_category('source-control')

def contains_git_repo(path: str) -> bool:
    """
    Returns true if the specified path
    contains a git based source control
    repository
    """

    return os.path.exists(os.path.join(path, '.git'))

def contains_svn_repo(path: str) -> bool:
    """
    Returns true if the specified path
    contains a svn based source control
    repository
    """

    __source_control_notify.warning('contains_svn_repo is not currently implemented!')

    return False

def get_git_local_head_revision(path: str) -> str:
    """
    Returns the local git head revision at the specified
    path if its a valid git repository
    """

    assert path != None
    assert path != ''

    revision = __default_revision
    try:
        import git
        repo = git.Repo(path)
        revision = repo.head.commit
    except Exception as e:
        __source_control_notify.warning('Failed to retrieve git head revision: %s' % str(e))

    return str(revision)

def get_svn_local_head_revision(path: str) -> str:
    """
    Returns the local svn head revision at the specified
    path if its a valid svn repository
    """

    assert path != None
    assert path != ''

    revision = __default_revision
    try:
        import pysvn
        svn_client = pysvn.Client()
        svn_info = svn_client.info(path)
        revision = svn_info['revision'].number
    except:
        __source_control_notify.warning('Failed to retrieve svn head revision: %s' % str(e))

    return str(revision)

def get_local_head_revision(path: str) -> str:
    """
    Returns the local head revision of the repository at the
    specified path. Supports svn and git repositories. Returns
    the value of default revision if the repository is not known
    or it is invalid
    """

    revision = __default_revision
    if contains_git_repo(path):
        revision = get_git_local_head_revision(path)
    elif contains_svn_repo(path):
        revision = get_svn_local_head_revision(path)

    return revision




"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.engine import runtime as __runtime
from panda3d_gemstone.logging import utilities as __logging
from panda3d.core import PStatCollector as __PStatCollector
from panda3d.core import PStatClient as __PStatClient

__notify = __logging.get_notify_category('performance')
__client = None

def get_profile_client() -> __PStatClient:
    """
    Returns the application's profiler client instance
    """

    global __client
    if __client is None:
        __client = __PStatClient()
    
    return __client

def is_profiling() -> bool:
    """
    Returns true if the application is currently being
    profiled
    """

    client = get_profile_client()
    return client.is_connected()

def connect_profiler(*args, **kwargs) -> bool:
    """
    Connects the application to a profiler server
    """

    if is_profiling():
        __notify.warning('Failed to connect profiler. A profiler is already connected')
        return False

    client = get_profile_client()
    client.connect(*args, **kwargs)

def disconnect_profiler() -> None:
    """
    Disconnects the application from a profiler server if it
    is currently connected
    """

    if not is_profiling():
        return

    client = get_profile_client()
    client.disconnect()

def toggle_profiling() -> None:
    """
    Toggles the current application profiling setting. Connecting
    to a server available at localhost
    """

    if is_profiling():
        connect_profiler()
    else:
        disconnect_profiler()

def resume_profiling_after_pause() -> None:
    """
    Resumes the profiling client after the simuilation has been paused for a while.
    This allows stats to continue exactly where it left off, instead of leaving a big gap that would represent
    a chug in the application's performance
    """

    client = get_profile_client()
    client.resume_after_pause()

def get_collector(label: str) -> __PStatCollector:
    """
    Creates and returns a new pstats collector instance
    """

    return __PStatCollector(label)

def __has_custom_collectors() -> bool:
    """
    Returns true if the custom_collectors dictionary
    is defined on our showbase instance
    """

    return hasattr(__runtime.base, 'custom_collectors')

def __has_custom_collector(name: str) -> bool:
    """
    Returns true if the custom collector currently
    exists
    """

    return name in __get_custom_collectors()

def __get_custom_collector(name: str) -> __PStatCollector:
    """
    Returns the custom collector if it exists
    """

    return __get_custom_collectors().get(name, None)

def __get_custom_collectors() -> list:
    """
    Returns a complete list of custom collectors
    """

    if not __has_custom_collectors():
        return []

    return __runtime.base.custom_collectors

def __create_custom_collector(name: str) -> object:
    """
    Creates and returns a custom PStatCollector
    instance
    """

    base = __runtime.base
    if not __has_custom_collectors():
        base.custom_collectors = {}

    if __has_custom_collector(name):
        __notify.warning('Attempted to create a new collector when it already exists! Name: %s' % name)
        return base.custom_collectors[collector_name]

    base.custom_collectors[collector_name] = __PStatCollector(collector_name)
    return base.custom_collectors[collector_name]

def stat_collection(func: object) -> object:
    """
    Wraps a function with a Panda3D PStatCollector object
    for timing its performance
    """

    collector_name = 'Debug:%s' % func.__name__
    if __has_custom_collector(collector_name):
        pstat = __get_custom_collector(collector_name)
    else:
        pstat = __create_custom_collector(collector_name)

    def do_pstat(*args, **kwargs) -> object:
        """
        Performs the timing operations for the
        wrapped function
        """

        pstat.start()
        results = func(*args, **kwargs)
        pstat.stop()

        return results

    do_pstat.__name__ = func.__name__
    do_pstat.__dict__ = func.__dict__
    do_pstat.__doc__ = func.__doc__

    return do_pstat

# Simplified decorator alias
pstat = stat_collection


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
from functools import reduce

from panda3d.core import BitMask32, Point2, Point3
from panda3d.core import Point4, Vec2, Vec3, Vec4

from panda3d_gemstone.framework.registry import ClassRegistry

class Caster(object):
    """
    Base class for all Gemstone framework casters
    """

    def accepted(self, value: object):
        """
        Returns true if the attempted cast is valid. Required to be implemented
        by child classes.
        """

        raise NotImplementedError('%s.accepted is not implemented!' % self.__class__.__name__)

    def __call__(self, value):
        """
        Returns the value of the attempted cast. Required to be implemented
        by the child classes.
        """

        raise NotImplementedError('%s called! Custom call function not implemented!' % self.__class__.__name__)

def get_class_registry() -> ClassRegistry:
    """
    Returns the class registry singleton object
    """

    return ClassRegistry.instantiate_singleton()

def register_caster(*args, **kwargs) -> None:
    """
    Registrs a class with the class registry
    """

    registry = get_class_registry()
    registry.register_class(is_caster=True, *args, **kwargs)

def __register_internal_caster(class_name) -> None:
    """
    Registers an internal caster class with the class registry
    based inside this module
    """

    package_path = 'panda3d_gemstone.framework.cast.%s' % class_name
    register_caster(class_name, package_path)

class CastBridge(object):
    """
    Bridge object to allow for object casting against
    known casters within the gemstone framework
    """

    __casters__ = []
    __primitives__ = [int, float]

    @classmethod
    def __attempt_primitive_cast(cls, type: object, value: object) -> bool:
        """
        Attempts to cast the value to a primitive type. Returns true and 
        the newly casted value on success. Otherwise False and NoneType
        """

        success = True
        output = None
        try:
            output = type(value)
        except (ValueError, TypeError):
            success = False

        return (success, output)

    def __new__(cls, value) -> object:
        """
        Attempts to create a new casted object
        from the supplied value on CastBridge 
        instantion
        """

        output = value

        # Attempt to cast against primtives
        for primitive in cls.__primitives__:
            success, output = cls.__attempt_primitive_cast(primitive, value)
            if success:
                return output

        # Attempt to cast against casters
        output = str(value).strip()
        for cast in cls.__casters__:
            if cast.accepted(output):
                output = cast(output)
                break

        return output

cast = CastBridge

class SequenceCaster(Caster):
    """
    """

    def __init__(self, cls: object, left: str, right: str, seperator: str):
        self._cls = cls
        self._left = left
        self._right = right
        self._seperator = seperator
    
    def accepted(self, value: str) -> bool:
        """
        """

        return value and value[0] == self._left and value[-1] == self._right

    def __split(self, value: str) -> list:
        """
        """

        words = []
        word = ''
        stack = []

        for i, v in enumerate(value):
            if v == self._left:
                stack.append(i)
                word += v
            elif v == self._right:
                stack.pop()
                word += v
            elif v == self._seperator and not stack:
                words.append(word.strip())
                word = ''
            else:
                word += v

        if word.strip():
            words.append(word.strip())
        
        return words

    def __call__(self, value):
        """
        """

        value = value[1:-1]
        values = self.__split(value)

        if values:
            values = [ cast(v) for v in values ]
            return self._cls(values)
        else:
            return self._cls()

class StringCaster(Caster):
    """
    Base caster for comparing the value of a cast to a string keyword
    and returning a basic type in the resulting cast
    """

    def __init__(self, key: str, obj: object):
        self._key = key
        self._obj = obj

    def accepted(self, value: str) -> bool:
        """
        Returns true if the attempted cast string value
        matches the string key of the caster
        """

        return value.lower() == self._key

    def __call__(self, value) -> object:
        """
        Returns the requested cast value provided
        by the child object
        """

        return self._obj

class ScalarCaster(Caster):
    """
    """

    def __init__(self, type_name: str, classes: []):
        self._type_name = type_name
        self._type_name_len = len(type_name)
        self._classes = classes

    def accepted(self, value: str) -> bool:
        """
        Returns true if the string value matches the scalar cast inputs
        """

        return value[:self._type_name_len] == self._type_name and value[self._type_name_len] == '(' and value[-1] == ')'

    def __call__(self, value: str) -> object:
        """
        """

        value = value[self._type_name_len + 1:-1]
        value = value.split(',')
        cls = self._classes[len(value) - 2]
        value = [float(n) for n in value]
        return cls(*value)

class FalseCast(StringCaster):
    """
    """

    def __init__(self):
        super().__init__('false', False)

__register_internal_caster('FalseCast')

class TrueCast(StringCaster):
    """
    """

    def __init__(self):
        super().__init__('true', True)

__register_internal_caster('TrueCast')

class ListCast(SequenceCaster):
    """
    """

    def __init__(self):
        super().__init__(list, '[', ']', ',')

__register_internal_caster('ListCast')

class NoneCast(StringCaster):
    """
    """

    def __init__(self):
        super().__init__('none', None)

__register_internal_caster('NoneCast')

class PandaBitCast(Caster):
    """
    """

    def __init__(self):
        self._type_name = 'Bit'
        self._type_name_len = len(self._type_name)

    def accepted(self, value: str) -> bool:
        """
        """

        return value[:self._type_name_len] == self._type_name and value[self._type_name_len] == '(' and value[-1] == ')'

    def __call__(self, value) -> object:
        """
        """

        value = value[self._type_name_len + 1:-1]
        value = value.split(',')
        bits = []

        for n in value:
            if n:
                n = int(n)
                bits.append(BitMask32.bit(n))

        if bits:
            return reduce(BitMask32.__or__, bits)
        else:
            return BitMask32()

__register_internal_caster('PandaBitCast')

class PandaPointCast(ScalarCaster):
    """
    """

    def __init__(self):
        super().__init__('Point', (Point2, Point3, Point4))

__register_internal_caster('PandaPointCast')

class PandaVectorCast(ScalarCaster):
    """
    """

    def __init__(self):
        super().__init__('Vec', (Vec2, Vec3, Vec4))

__register_internal_caster('PandaVectorCast')

class QuotedStringCast(Caster):
    """
    """

    def accepted(self, value: str) -> bool:
        """
        """

        return value and value[0] == value[-1] and value[0] in ('"', "'")

    def __call__(self, value: str) -> object:
        """
        """

        return value[1:-1]

__register_internal_caster('QuotedStringCast')

class TupleCast(SequenceCaster):
    """
    """

    def __init__(self):
        super().__init__(tuple, '(', ')', ',')

__register_internal_caster('TupleCast')

def __setup_casting() -> None:
    """
    Performs casting setup operations
    """

    # Setup bridge casters from registry
    CastBridge.__casters__ = [cls() for cls in get_class_registry().query_meta(is_caster=True)]

# Perform setup on import
__setup_casting()


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from six import with_metaclass

from panda3d.core import ConfigVariable, ConfigVariableList, ConfigVariableString
from panda3d.core import ConfigVariableFilename, ConfigVariableBool, ConfigVariableInt
from panda3d.core import ConfigVariableDouble, ConfigVariableColor, ConfigVariableInt64
from panda3d.core import ConfigVariableSearchPath, ConfigFlags
from panda3d.core import load_prc_file as _load_prc_file
from panda3d.core import load_prc_file_data as _load_prc_file_data

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.io.file_system import path_exists
from panda3d_gemstone.logging.utilities import get_notify_category as __get_notify_category

__prc_notify = __get_notify_category('prc')
__prc_notify.setInfo(True)

def load_prc_file_data(data: str, label: str = '') -> None:
    """
    Loads the requested string of PRC data into the Panda3D
    runtime configuration under the requested label if 
    provided. Otherwise empty string
    """

    assert data != None
    assert label != None

    # Check if the base has already been defined
    # if it has been defined warn the user.
    if runtime.has_base():
        __prc_notify.warning('Showbase has already been defined. PRC changes may be ignored')

    if label != '' and label.isspace() == False:
        __prc_notify.info('Setting PRC data for label: %s' % label)
    _load_prc_file_data(label, data)

def load_prc_file(path: str, optional: bool = False) -> bool:
    """
    Attempts to load a prc file into the application. Returns
    True if succesful otherwise False
    """

    # Check if the base has already been defined
    # if it has been defined warn the user.
    if runtime.has_base():
        __prc_notify.warning('Showbase has already been defined. PRC changes may be ignored')

    if not path_exists(path) and not optional:
        __prc_notify.error('Failed to load prc file: %s. File does not exist' % path)
        return False
    
    # Return if the path does not exist and we are optional
    if not path_exists(path) and optional:
        __prc_notify.warning('Skipping optional prc: %s' % path)
        return False

    if optional:
        __prc_notify.info('Loading optional runtime config: %s' % path)
    else:
        __prc_notify.info('Loading runtime config: %s' % path)
    _load_prc_file(path)
    return True

def get_prc_list(key: str) -> list:
    """
    Retrieves a int variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    return ConfigVariableList(key)

def get_prc_search_path(key: str, default: int = 0) -> object:
    """
    Retrieves a search path variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableSearchPath(key, default)
    return variable.get_value()

def get_prc_color(key: str, default: int = 0) -> object:
    """
    Retrieves a color variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableColor(key, default)
    return variable.get_value()

def get_prc_filename(key: str, default: int = 0) -> object:
    """
    Retrieves a filename variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableFilename(key, default)
    return variable.get_value()

def get_prc_double(key: str, default: int = 0) -> float:
    """
    Retrieves a double variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableDouble(key, default)
    return variable.get_value()

def get_prc_int64(key: str, default: int = 0) -> int:
    """
    Retrieves a int64 variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableInt64(key, default)
    return variable.get_value()

def get_prc_int(key: str, default: int = 0, list_result: bool = False) -> object:
    """
    Retrieves a int variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    if default is None:
        default = 0

    variable = ConfigVariableInt(key, default)
    num_words = variable.get_num_words()
    if num_words > 0 and list_result:
        return [variable.get_word(i) for i in range(num_words)]
    else:
        return variable.get_value()

def get_prc_string_value(key: str, default: str = '') -> str:
    """
    Retrieves a string variable from the Panda3D
    runtime configuration if present using the
    get_string_value method. Otherwise returns 
    the default value
    """

    variable = ConfigVariableString(key)
    value = variable.get_string_value()
    if not value:
        value = default

    return value

def get_prc_string(key: str, default: str = '') -> str:
    """
    Retrieves a string variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableString(key, default)
    return variable.get_value()

def get_prc_bool(key: str, default: bool = False) -> bool:
    """
    Retrieves a boolean variable from the Panda3D
    runtime configuration if present. Otherwise
    the default value
    """

    variable = ConfigVariableBool(key, default)
    return variable.get_value()

def get_prc_value_type(key: str) -> object:
    """
    Returns the value type of the Panda3D runtime configuration
    key if it exists
    """

    return ConfigVariable(key).get_value_type()

def set_prc_string_value(key: str, value: str) -> None:
    """
    Sets the requested string key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariable(key).set_string_value(value)

def set_prc_string(key: str, value: str) -> None:
    """
    Sets the requested string key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableString(key).set_value(value)

def set_prc_filename(key: str, value: object) -> None:
    """
    Sets the requested filename key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableFilename(key).set_value(value)

def set_prc_bool(key: str, value: bool) -> None:
    """
    Sets the requested boolean key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableBool(key).set_value(value)

def set_prc_int(key: str, value: int) -> None:
    """
    Sets the requested int key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableInt(key).set_value(value)

def set_prc_double(key: str, value: float) -> None:
    """
    Sets the requested double/float key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableDouble(key).set_value(value)

def set_prc_search_path(key: str, value: str) -> None:
    """
    Sets the requested search path key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableSearchPath(key).set_value(value)

def set_prc_int64(key: str, value: str) -> None:
    """
    Sets the requested int64 key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableInt64(key).set_value(value)

def set_prc_color(key: str, value: str) -> None:
    """
    Sets the requested color key and value inside the Panda3D runtime
    configuration.
    """

    ConfigVariableColor(key).set_value(value)

__prc_type_map = {
    ConfigVariable.VT_undefined: (get_prc_string_value, set_prc_string_value),
    ConfigVariable.VT_list: (get_prc_list, None),
    ConfigVariable.VT_string: (get_prc_string, set_prc_string),
    ConfigVariable.VT_filename: (get_prc_filename, set_prc_filename),
    ConfigVariable.VT_bool: (get_prc_bool, set_prc_bool),
    ConfigVariable.VT_int: (get_prc_int, set_prc_int),
    ConfigVariable.VT_double: (get_prc_double, get_prc_double),
    ConfigVariable.VT_enum: (get_prc_string_value, set_prc_string_value),
    ConfigVariable.VT_search_path: (get_prc_search_path, set_prc_search_path),
    ConfigVariable.VT_int64: (get_prc_int64, set_prc_int64),
    ConfigVariable.VT_color: (get_prc_color, set_prc_color)
}

def get_prc_value(key: str, default: object = None) -> object:
    """
    Attempts to retrieve the Panda3D runtime configuration key
    using its predicted value type. Returns the retrieved
    value or default if its invalid or not found
    """

    value_type = get_prc_value_type(key)
    type_info = __prc_type_map.get(value_type, None)
    if type_info is None:
        raise ValueError('PRC type (%s) is not a valid prc value type.' % value_type)

    getter, setter = type_info
    if getter is None:
        __prc_notify.warning('PRC type (%s) does not support value retrieval' % value_type)
        return

    value = getter(key, default)
    return value

def set_prc_value(key: str, value: object) -> None:
    """
    Sets the value of the Panda3D runtime configuration key
    if its supported by the value type
    """

    value_type = get_prc_value_type(key)
    type_info = __prc_type_map.get(value_type, None)
    if type_info is None:
        raise ValueError('PRC type (%s) is not a valid prc value type.' % value_type)

    getter, setter = type_info
    if setter is None:
        __prc_notify.warning('Failed to set key: %s. PRC type (%s) does not support value setting' % (key, value_type))
        return

    setter(key, value)

def has_prc_key(key: str) -> bool:
    """
    Returns true if the requested key was defined in the Panda3d
    engine configuration
    """

    return ConfigVariable(key).has_value()

def get_config_manager() -> object:
    """
    Returns the Panda3D ConfigVariableManager object
    """

    from panda3d.core import ConfigVariableManager
    return ConfigVariableManager.get_global_ptr()

class MetaConfig(type):
    """
    Metaclass for the Panda3D runtime configuration. 
    Allows use of __getitem__ and __setitem__ to access the engines 
    runtime configuration.
    """

    def __getitem__(cls, key: str) -> object:
        """
        Custom get item handler for allowing index access
        to the Panda3D runtime configuration
        """

        return get_prc_value(key)

    def __setitem__(cls, key: str, value: object) -> None:
        """
        Custom set item handler for allowing index access
        to the Panda3D runtime configuration
        """

        set_prc_value(key, value)

    def __contains__(cls, key: str) -> bool:
        """
        Custom contains handler for allowing index access
        to the Panda3D runtime configuration
        """

        return has_prc_key(key)

class RuntimeConfig(with_metaclass(MetaConfig, object)):
    """
    Runtime config access wrapper for the Panda3D
    runtime configuration
    """

class DConfigMixIn(object):
    """
    MixIn class for adding support for the original
    legacy Panda3D DCConfig class inside direct
    """

    def GetBool(self, sym: str, default: bool = False) -> bool:
        """
        Retrieves a boolean value from the prc configuration
        """

        return ConfigVariableBool(sym, default, "DConfig", ConfigFlags.F_dconfig).value

    def GetInt(self, sym: str, default: int = 0) -> int:
        """
        Retrieves an integer value from the prc configuration
        """

        return ConfigVariableInt(sym, default, "DConfig", ConfigFlags.F_dconfig).value

    def GetDouble(self, sym: str, default: float = 0.0) -> float:
        """
        Retrieves a double/float value from the prc configuration
        """

        return ConfigVariableDouble(sym, default, "DConfig", ConfigFlags.F_dconfig).value

    def GetString(self, sym: str, default: str = "") -> str:
        """
        Retrieves a string value from the prc configuration
        """

        return ConfigVariableString(sym, default, "DConfig", ConfigFlags.F_dconfig).value

    GetFloat = GetDouble

class ShowbaseConfig(RuntimeConfig, DConfigMixIn):
    """
    Custom RuntimeConfig object that doubles as a replacement
    for the legacy Panda3D DConfig object for replacing the config
    variable inside ShowBase and keeping legacy direct module
    support
    """


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import sys
import os

import glob

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.cast import cast
from panda3d_gemstone.framework.utilities import disallow_production
from panda3d_gemstone.io.file_system import get_file_date, fix_path
from panda3d_gemstone.engine.messenger import MessageListener
from panda3d_gemstone.logging import utilities as logging

from panda3d.core import LoaderOptions, NodePath

from functools import reduce

__resource_notify = logging.get_notify_category('resource')

class ExportProperty(InternalObject):
    """
    """

    FILENAME_RE = re.compile('^\\s*([^\t\n\r\x0c\x0b \\(]+)\\s*(.*)$')
    OPTION_RE = re.compile('^\\s*(?:(\\w+)\\(' + '([^\\(]*)' + '\\))\\s*(.*)$')

    def __init__(self, key: object = None, data: object = None, converter: object = None):
        super().__init__()
        self.output_filename = key
        self.converter = converter
        self.input_filename = ''
        self.options = {}

        if data:
            self.parse(data)

    def generate_post_egg2bam(self, egg2bam_options: object = None) -> object:
        """
        """

        binary = ExportProperty(converter='Egg2Bam')
        binary.input_filename = self.output_filename
        binary.output_filename = binary.input_filename

        if egg2bam_options:
            binary.options = egg2bam_options

        return binary

    def generate_post_bam2pz(self) -> object:
        """
        """

        compress = ExportProperty(converter='Bam2Pz')
        compress.input_filename = self.output_filename
        compress.output_filename = compress.input_filename

        return compress

    def generate_post_egg2bampz(self, egg2bam_options: object = None) -> object:
        """
        """

        binary = self.generate_post_egg2bam(egg2bam_options)
        compress = binary.generate_post_egg2bampz()

        return (binary, compress)

    def parse(self, data: dict) -> None:
        """
        Attempts to parse the incoming data dictionary into options
        """

        m = ExportProperty.FILENAME_RE.match(data)
        if not m:
            self.notify.error('ResourceProperty: syntax error (Failed to parse data. Naming incorrect)')
            return
        
        self.input_filename = m.group(1)
        rest = m.group(2)

        while rest:
            m = ExportProperty.OPTION_RE.match(rest)
            if not m:
                self.notify.error('ResourceProperty: syntax error (Failed to parse data. Options format incorrect)')
                rest = ''
                continue

                optname = '-' + m.group(1)
                optval = cast(m.group(2))
                rest = m.group(3)
                self.options[optname] = optval

    def get_option_string(self) -> str:
        """
        """

        option_str = ''
        for optname, optvalue in list(self.options.items()):
            option_str = option_str + ' %s %s' % (optname, optvalue)

        return option_str

    def __str__(self):
        """
        Custom string formatter for the ExportProperty class
        """
        
        label  = '%s <inputFile: %s, outputFile: %s, converter: %s, options: %s>' % (
            self.__class__.__name__, self.input_filename, self.output_filename, self.converter, self.options)

        return label

class Resource(Configurable):
    """
    Represents a Gemstone framework/Panda3D resource
    """

    def __init__(self, config_path: str, force_export: bool = False, load_immediate: bool = True):
        self._export_failed = True
        self._export_steps = []
        super().__init__(config_path)
        self._skip_export = self.pop('skipExport', False)
        self._loaded = False
        self.initialize()

        if load_immediate:
            self._load(force_export)
        else:
            self.reset_export_steps()
            self.export(force_export)

    @property
    def loaded(self) -> bool:
        """
        Returns the loaded state
        """

        return self._loaded

    @property
    def export_failed(self) -> bool:
        """
        Returns the export failed state
        """

        return self._export_failed

    def add_export_step(self, export_step: object) -> None:
        """
        Adds a new export step to the Resource
        """

        self._export_steps.append(export_step)

    def extend_export_steps(self, export_steps: list) -> None:
        """
        Extends the Resource's export steps with an export steps list
        """

        #TODO: fix this
        #self._export_steps.extend(export_steps)

    def get_export_step(self, index: int) -> object:
        """
        Retrieves the export step at index
        """

        return self._export_steps[index]

    def reset_export_steps(self) -> None:
        """
        Resets the Resources's export steps
        """

        self._export_steps = []

    def copy_to(self, new_parent: NodePath, sort: int = 0) -> NodePath:
        """
        Copys the resource object to a new NodePath instance
        """

        model = self.__class__(self.path)
        model.reparent_to(new_parent, sort)

        return model

    def _load(self, force_export: bool) -> bool:
        """
        Loads the game resource
        """

        # Verify we are not already loaded
        if self._loaded:
            return False

        self.reset_export_steps()
        self.export(force_export)
        self._loaded = True

        return True

    def _unload(self) -> bool:
        """
        Unloads the resource object
        """

        # Verify we are currently loaded
        if not self._loaded:
            return False

        self._loaded = False
        return True

    def reload(self, force_reload: bool = False, force_export: bool = False) -> None:
        """
        Reloads the resource object
        """

        if force_reload and self._loaded:
            self._unload()
        self.load()

        if force_reload:
            self.__load(force_export)
        self.initialize()

    def export(self, force_export: bool) -> None:
        """
        Exports the resource using the export steps
        """

        self.reset_export_steps()
        self._export_failed = False

    @disallow_production
    def attempt_reload(self, file_path: str) -> None:
        """
        Called by the ResourceManager when an ApplicationReload event is fired
        to attempt to reload the Resource. Intended to be overriden by children
        that support reloading
        """

    @disallow_production
    def supports_reloading(self) -> bool:
        """
        Returns true if the object supports reloading. Intended to be overriden by children
        """

        return True

def __partition_export_steps(export_steps: list) -> list:
    """
    """

    chains = []
    if len(export_steps) == 0:
        return chains

    current_chain = [export_steps[0]]
    for export_step in export_steps[1:]:
        if export_step.input_filename == current_chain[-1].output_filename:
            current_chain.apppend(export_step)
        else:
            chains.append(current_chain)
            current_chain = [export_step]

    if len(current_chain) > 0:
        chains.append(current_chain)

    return chains

def __check_for_necessary_export(chains: list, ini_file_date: object) -> tuple:
    """
    """

    result = True
    remaining_steps = []
    force_export = False
    for chain in chains:
        if len(chain) > 0:
            output_file = __get_output_filename(chain[-1])
            if output_file:
                output_date = get_file_date(output_file)
                if output_date < ini_file_date:
                    force_export = True

    if not force_export:
        for chain in chains:
            if len(chain) > 0:
                source_file = chain[0].input_filename
                output_file = __get_output_filename(chain[-1])
                if source_file and output_file:
                    source_date = get_file_date(source_file)
                    output_date = get_file_date(output_date)

                    if source_date > output_date:
                        remaining_steps.extend(chain)
                else:
                    result = False
            else:
                __resource_notify.warning('__check_for_necessary_export: Found an empty export chain!')
    else:
        remaining_steps = reduce(list.__add__, chains)

    return (result, remaining_steps, force_export)

def __get_output_filename(export_property: ExportProperty) -> str:
    """
    """

    filename = None
    export_info = EXPORTINFO.get(export_property.converter, None)
    if export_info:
        export_func, input_extensions, output_extension = export_info
        root, ext = os.path.splitext(export_property.output_filename)
        filename = root + output_extension
    else:
        __resource_notify.warning('__get_output_filename: Export lookup for %s failed' % export_property.converter)

    return filename

def __build_in_out_names(export_step: object, input_extensions: object, output_extension: object) -> tuple:
    """
    """

def __export_if_source_newer(converter: object, source_file: str, target_file: str, options: str = '', force_export: bool = False) -> tuple:
    """
    """

def __export(converter: object, input_filename: str, output_filename: str, options: str = '', callformat: str = '%s %s - %s %s') -> object:
    """
    """

def __export_func_generic(converter: object, export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> tuple:
    """
    """

    in_file, out_file = __build_in_out_names(export_step, input_extensions, output_extension)
    if in_file and out_file:
        return __export_if_source_newer(converter, in_file, out_file, export_step.get_option_string(), force_export)
    else:
        return (False, False)

def export_func_mayaToEgg(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a maya2egg export operation
    """

    return __export_func_generic('maya2egg', export_step, force_export, input_extensions, output_extension)

def export_func_eggToBam(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a egg2bam export operation
    """

    return __export_func_generic('egg2bam', export_step, force_export, input_extensions, output_extension)

def export_func_bamToBamPz(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a pzip export operation
    """

    return __export_func_generic('pzip', export_step, force_export, input_extensions, output_extension)

def export_func_opt_char(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a egg-optchar export operation
    """

    return __export_func_generic('egg-optchar', export_step, force_export, input_extensions, output_extension)

def export_func_make_font(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a egg-mkfont export operation
    """

    return __export_func_generic('egg-mkfont', export_step, force_export, input_extensions, output_extension)

def export_func_texture_cards(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a egg-texture-cards export operation
    """

    #TODO: write this
    raise NotImplementedError('export_func_texture_cards is not currently implemented!')

def export_func_png_to_ico(export_step: object, force_export: bool, input_extensions: object, output_extension: object) -> object:
    """
    Performs a png2ico export operation
    """

    in_file, out_file = __build_in_out_names(export_step, input_extensions, output_extension)
    if in_file and out_file:
        return __export_if_source_newer(converter, in_file, out_file, '', force_export, '%s %s %s %s')
    else:
        return (False, False)

EXPORTINFO = {
    'Export': (export_func_mayaToEgg, ('.mb', '.ma'), '.egg'),
    'Egg2Bam': (export_func_eggToBam, ('.egg',), '.bam'),
    'Bam2Pz': (export_func_bamToBamPz, ('.bam',), '.bam.pz'),
    'TextureCards': (export_func_texture_cards, ('.jpg', '.bmp', '.tga', '.png'), '.egg'),
    'OptChar': (export_func_opt_char, ('.egg',), '.egg'),
    'MakeFont': (export_func_make_font, ('.ttf',), '.egg'),
    'Png2Ico': (export_func_png_to_ico, ('.png',), '.ico')
}

CONVERTER = {
    'win32': {
        'maya2egg': 'maya2egg2008.exe',
        'egg2bam': 'egg2bam.exe',
        'pzip': 'pzip.exe',
        'egg-texture-cards': 'egg-texture-cards.exe',
        'egg-optchar': 'egg-optchar.exe',
        'egg-mkfont': 'egg-mkfont.exe',
        'png2ico': 'png2ico.exe'
    },
    'darwin': {
        'maya2egg': 'maya2egg2008',
        'egg2bam': 'egg2bam',
        'pzip': 'pzip',
        'egg-texture-cards': 'egg-texture-cards',
        'egg-optchar': 'egg-optchar',
        'egg-mkfont': 'egg-mkfont',
        'png2ico': 'png2ico'
    },
    'linux2': {
        'maya2egg': 'maya2egg2008',
        'egg2bam': 'egg2bam',
        'pzip': 'pzip',
        'egg-texture-cards': 'egg-texture-cards',
        'egg-optchar': 'egg-optchar',
        'egg-mkfont': 'egg-mkfont',
        'png2ico': 'png2ico'
    }
}

class ResourceManager(Singleton, MessageListener, Configurable):
    """
    Singleton instance for managing resources inside Gemstone
    """

    def __init__(self, config_path: str = None):
        Singleton.__init__(self)
        MessageListener.__init__(self)
        self.accept('gs-file-changed', self.__on_file_changed)
        self._skip_export = False
        self._resources = []
        self._loader_options = LoaderOptions()
        Configurable.__init__(self, config_path)

    def destroy(self) -> None:
        """
        Destroys the resource manager instance
        """

        Singleton.destroy(self)
        MessageListener.destroy(self)

    @property
    def skip_export(self) -> bool:
        """
        Property access to the get_skip_exports
        method
        """

        return self.get_skip_export()

    @skip_export.setter
    def skip_export(self, enable: bool) -> None:
        """
        Setter access to the set_skip_exports
        method
        """

        self.set_skip_export(enable)

    @property
    def loader_options(self) -> LoaderOptions:
        """
        Property access to get_loader_options
        method
        """

        return self.get_loader_options()

    @loader_options.setter
    def loader_options(self, flags: object) -> None:
        """
        Setter access to the set_loader_options
        method
        """

        self.set_loader_options(flags)

    def register(self, resource: Resource) -> None:
        """
        Registers the requested resource with the resource manager.
        """

    def unregister(self, resource: Resource) -> None:
        """
        Unregisters the requested resource from the resource manager.
        """

    def reload(self, force_reload: bool = False, force_export: bool = False) -> None:
        """
        Reloads the resources currently attached to the resource manager
        """

        for resource in self._resources:
            resource.reload(force_reload, force_export)

    @disallow_production
    def __on_file_changed(self, file_path: str, ext: str) -> None:
        """
        Called on ApplicationReloader file change event.
        """

        for resource in self._resources:
            if resource.supports_reloading():
                resource.attempt_reload(file_path)

    def set_skip_export(self, enable: bool) -> None:
        """
        Sets the resource manager's skip export flag
        """

        self._skip_export = enable

    def get_skip_export(self) -> None:
        """
        Retrieves the resource manager's skip export flag
        """

        return self._skip_export

    def get_loader_options(self) -> LoaderOptions:
        """
        Retrieves the resource manager's loader flags
        """

        return self._loader_options
    
    def set_loader_options(self, flags: object) -> None:
        """
        Sets the resource manager's loader flags
        """

        self._loader_options.set_flags(flags)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d.core import NodePath, TextNode
from panda3d.core import Texture

from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.resource import Resource, ExportProperty

from panda3d_gemstone.engine.model_utilities import get_model_texture_objects

class Text(NodePath, Configurable, InternalObject):
    """
    Gemstone framework wrapper for the Panda3D TextNode object
    for displaying text in the scene graph
    """

    Alignment = {
        'left': TextNode.ALeft,
        'right': TextNode.ARight,
        'center': TextNode.ACenter,
        TextNode.ALeft: TextNode.ALeft,
        TextNode.ARight: TextNode.ARight,
        TextNode.ACenter: TextNode.ACenter
    }

    def __init__(self, config_path: str, section: str = 'Configuration'):
        NodePath.__init__(self, '')
        Configurable.__init__(self, config_path, section)

        self.__text_node = TextNode('TextBillboard')
        self.assign(self.attach_new_node(self.__text_node))
        self.set_billboard_point_eye()
        self.set_light_off()
        self.set_color_off()
        self.set_scale(0.0425)
        self.set_pos(0, 0, 0)
        self.set_color(1.0, 1.0, 1.0, 1.0)
        self.set_shadow_color(0.0, 0.0, 0.0, 1.0)
        self.set_shadow_offset(0.08, 0.08)
        self.set_text_alignment('center')
        self.set_text('')
        self.initialize()

    def set_enabled(self, enabled: bool) -> None:
        """
        Sets the Text object's enabled state
        """

        if enabled:
            self.show()
        else:
            self.hide()

    def get_shadow_color(self) -> object:
        """
        Returns the Text objects's shadow color
        """

        return self.__text_node.get_shadow_color()

    def set_shadow_color(self, *color) -> None:
        """
        Sets the Text object's shadow color
        """

        self.__text_node.set_shadow_color(*color)

    def get_color(self) -> object:
        """
        Returns this Text object's text color
        """

        return self.__text_node.get_text_color()

    def set_color(self, *color) -> None:
        """
        Sets this Text object's text color
        """

        self.__text_node.set_text_color(*color)

    def get_shadow_offset(self) -> object:
        """
        Returns this Text object's shadow offset
        """

        return self.__text_node.get_shadow()

    def set_shadow_offset(self, *offset) -> None:
        """
        Sets thie Text object's shadow offset
        """

        self.__text_node.set_shadow_color(*offset)

    def clear_shadow_color(self) -> None:
        """
        Clears the text object's shadow color
        """

        self.__text_node.clear_shadow_color()

    def clear_shadow(self) -> None:
        """
        Clears the text object's shadow
        """

        self.__text_node.clear_shadow()

    def set_font(self, font_path: str) -> None:
        """
        Sets the text object's font resource
        """

        assert font_path != None
        assert font_path != ''

        font = Font(font_path)
        if not font:
            self.notify.warning('Failed to set font. Could not open font "%s"' % font_path)
            return
        
        text_font = font.get_text_font()
        self.__text_node.set_font(text_font)

    def set_text(self, text) -> None:
        """
        Sets the text object's displayed text
        """

        self.__text_node.set_text(text)

    def get_text(self) -> str:
        """
        Gets the text object's displayed text
        """

        return self.__text_node.get_text()

    def get_text_alignment(self) -> object:
        """
        Returns the text object's text alignment
        """

        return self.__text_node.get_align()

    def get_text_alignment_name(self, align: object) -> str:
        """
        Returns the text alignment name from instance
        """

        for k, v in list(Text.Alignment.items()):
            if v == align and isinstance(k, str):
                return k
        
    def set_text_alignment(self, align: object) -> None:
        """
        Sets the text object's alignment property
        """

        if isinstance(align, str):
            align = align.lower()

        if not align in Text.Alignment:
            self.notify.warning('Failed to set alignment. Invalid alignement "%s" specified' % align)
            return
        
        self.__text_node.set_align(Text.Alignment[align])

class Font(Resource, InternalObject):
    """
    Represents a Font resource in the Gemstone framework. Wrapping the
    native Panda3D font objects
    """

    def __init__(self, config_path: str, force_export: bool = False):
        self.__name = ''
        self.__fonts = []
        self.__font_data = {}
        self.__text_font = None
        self.__line_height = None
        self.__font_textures = []
        self.set_disable_filtering(False)

        Resource.__init__(self, config_path, force_export)
        InternalObject.__init__(self)

    @property
    def text_font(self) -> object:
        """
        Returns the font resource as a Panda3D text font object
        through a property
        """

        return self.get_text_font()

    def get_text_font(self) -> object:
        """
        Returns the font resource as a Panda3D text font object
        """

        return self.__text_font

    def load_font_data(self, data: dict) -> None:
        """
        Loads the font data dictionary into the resource 
        object
        """

        self.__font_data = data

    def set_line_height(self, data: object) -> None:
        """
        Sets the font resource's line height
        """

        self.__line_height = data

    def disable_texture_filtering(self, disable_filtering: bool) -> None:
        """
        Sets the text filtering state as either true or false
        """

        if disable_filtering:
            mag_filter = Texture.FTNearest
        else:
            mag_filter = Texture.FTLinear

        for tex in self.__font_textures:
            tex.set_mag_filter(mag_filter)

    def export(self, force_export: bool) -> None:
        """
        Attempts to export the font resource
        """

        if len(self.__font_data) == 1:
            params = list(list(self.__font_data.items())[0])
            params.append('MakeFont')
            export = ExportProperty(*params)
            self.add_export_step(export)
            self.add_export_step(export.generate_post_egg2bam)
            self.__fonts.append(self.get_export_step(-1).output_filename)
            Resource.export(self, force_export)
        elif len(self.__font_data) > 1:
            self.notify.error('%s does not support multipart models' % self.__class__.__name__)

    def set_name(self, name: str) -> None:
        """
        Sets the font resource's name
        """

        self.__name = name

    def get_name(self) -> str:
        """
        Retrieves the font resource's name
        """

        return self.__name

    def _load(self, force_export: bool) -> bool:
        """
        Loads the font resource into memory
        """

    def _unload(self) -> None:
        """
        Unloads the font resource from memory
        """

        if self.loaded:

            if len(self.__fonts):
                self.notify.warning('Font unloading not yet implemented. Attempted to unload font: %s' % self.__fonts[0])
            return True

        return False

class TextBillboard(Configurable, InternalObject):
    """
    Represents a billboard text node in the Gemstone framework
    """

    Alignment = {
        'left': TextNode.ALeft,
        'right': TextNode.ARight,
        'center': TextNode.ACenter,
        TextNode.ALeft: TextNode.ALeft,
        TextNode.ARight: TextNode.ARight,
        TextNode.ACenter: TextNode.ACenter
    }

    def __init__(self, config_path: str, parent: object):
        InternalObject.__init__(self)
        self.__text_node = TextNode(self.__class__.__name__)
        self.__nodepath = parent.attach_new_node(self.__text_node)
        self.__nodepath.set_billboard_point_eye()
        self.__nodepath.set_light_off()
        self.__nodepath.set_color_off()
        self.set_enabled(True)
        self.set_color(Vec4(1, 1, 1, 1))
        self.set_align(TextNode.ACenter)
        self.set_text_size(0.7)
        self.set_position(Vec3(0, 0, 4.2))
        self.set_font(None)
        self.set_text('')
        
        Configurable.__init__(self, config_path)
        self.initialize()

    def set_enabled(self, enabled: bool) -> None:
        """
        Sets the object's enabled visibility state
        """

        if enabled:
            self.__nodepath.show()
        else:
            self.__nodepath.hide()

    def set_color(self, color: object) -> None:
        """
        Sets the object's text color
        """

        self.__text_node.set_text_color(color)

    def get_color(self) -> object:
        """
        Returns the object's text color
        """

        return self.__text_node.get_text_color()

    def set_shadow_color(self, color: object) -> None:
        """
        Sets the object's shadow color
        """

        self.__text_node.set_shadow_color(color)

    def get_shadow_color(self) -> object:
        """
        Returns the object's shadow color
        """

        return self.__text_node.get_shadow_color()

    def set_font(self, font: object) -> None:
        """
        Sets the billboard's font object
        """

        if not font:
            return

        f = Font(font)
        tf = f.get_text_font()
        if tf:
            self.__text_node.set_font(tf)

    def set_text(self, text: str) -> None:
        """
        Sets the billboard's text value
        """

        self.__text_node.set_text(text)

    def get_text(self) -> str:
        """
        Returns this billboard's text value
        """

        return self.__text_node.get_text()

    def set_shadow_offset(self, offset: object) -> None:
        """
        Sets the text shadow offset
        """

        self.__text_node.set_shadow(offset)

    def get_shadow_offset(self) -> object:
        """
        Returns the text shadow offset
        """

        return self.__text_node.get_shadow(offset)

    def set_align(self, align: object) -> None:
        """
        Sets the object's text alignment
        """

        a = TextBillboard.Alignment.get(align.lower(), None)
        if not a:
            self.notify.warning('Failed to align. Invalid alignment specified: %s' % str(align))

        self.__text_node.set_align(a)

    def reparent_to(self, nodepath: object) -> None:
        """
        Reparents the object to the node path
        """

        self.__nodepath.reparent_to(nodepath)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from direct.showbase.ShowBase import ShowBase

from panda3d_gemstone.editor.gizmos.grid import GizmoGrid
from panda3d_gemstone.engine.showbase import ApplicationShowBase
from panda3d_gemstone.logging import utilities as logging

class EditorShowbase(ApplicationShowBase):
    """
    Custom Showbase instance for the Gemstone editor
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_window_title(self, *args, **kwargs) -> None:
        """
        Setup the window title for the ShowBase instance.
        Not used by this instance of ShowBase
        """

    def oobe(self, cam: object = None) -> None:
        """
        Enable a special "out-of-body experience" mouse-interface
        mode.  This can be used when a "god" camera is needed; it
        moves the camera node out from under its normal node and sets
        the world up in trackball state.  Button events are still sent
        to the normal mouse action node (e.g. the DriveInterface), and
        mouse events, if needed, may be sent to the normal node by
        holding down the Control key.
        This is different than useTrackball(), which simply changes
        the existing mouse action to a trackball interface.  In fact,
        OOBE mode doesn't care whether useDrive() or useTrackball() is
        in effect; it just temporarily layers a new trackball
        interface on top of whatever the basic interface is.  You can
        even switch between useDrive() and useTrackball() while OOBE
        mode is in effect.
        This is a toggle; the second time this function is called, it
        disables the mode.
        """

        self.notify.warning('OOBE is not supported by the Gemstone editor application')


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.logging.utilities import get_notify_category

_notify = get_notify_category('singleton')

class Singleton(object):
    """
    Represents a singleton object in the gemstone framework
    """

    _singleton_instance = None

    def __init__(self):
        if not self.__class__.is_instantiated():
            self.__class__.reset_singleton(self)
        else:
            raise RuntimeError('%s singleton already exists' % self.__class__.__name__)
        self.setup()

    @classmethod
    def instantiate_singleton(cls, *args, **kwargs):
        """
        Instantiates a new singleton instance
        """

        if not cls.is_instantiated():
            cls.reset_singleton(cls(*args, **kwargs))

        return cls.get_singleton()

    @classmethod
    def reset_singleton(cls, inst: object = None) -> None:
        """
        Resets the singleton instance
        """

        # Destroy existing singleton if one exists
        if cls._singleton_instance is not None and id(inst) != id(cls._singleton_instance):
            cls.destroy(cls._singleton_instance)

        # Setup new singleton
        cls._singleton_instance = inst

    @classmethod
    def get_singleton(cls, silent: bool = False) -> object:
        """
        Returns the current singleton instance
        """

        instance = cls._singleton_instance
        if not instance and not silent:
            _notify.warning('Failed to get singleton. %s is not instantiated' % (
                cls.__name__))

        return instance

    @classmethod
    def is_instantiated(cls) -> bool:
        """
        Returns true if the current singleton has an instance
        """

        return cls._singleton_instance != None

    def setup(self) -> None:
        """
        Performs setup operations on the singleton
        """

    def destroy(self) -> None:
        """
        Performs shutdown operations on the singleton
        """

class SingletonWrapper(Singleton):
    """
    Custom Singleton instance for wrapping Python objects as Gemstone
    singleton objects.
    """

    WRAPPED_OBJECT = None

    def __init__(self, *args, **kwargs):
        assert self.WRAPPED_OBJECT != None
        self.__object = self.WRAPPED_OBJECT(*args, **kwargs)
        assert self.__object != None

    @property
    def wrapped(self) -> object:
        """
        Singleton's wrapped object instance
        """

        return self.get_wrapped()

    def get_wrapped(self) -> object:
        """
        Returns the wrapped object instance
        """

        return self.__object

    def __getattr__(self, key: str) -> object:
        """
        Custom attributes retrieval handler to pass through unknown
        attributes to our wrapped object instance
        """

        result = None
        if self.__object != None:
            if hasattr(self.__object, key):
                result = getattr(self.__object, key)

        if not result:
            raise AttributeError('%s instance does not have attribute %s' % (
                self.__class__.__name__, key))    

        return result


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging
import importlib

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.internal_object import InternalObject

class ClassRegistry(Singleton, InternalObject):
    """
    Singleton registry containing all classes used for networking, configuration
    and casting within the gemstone framework.
    """

    def __init__(self):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self._classes = {}

    @property
    def classes(self) -> dict:
        """
        Getter for retrieving the registry
        classes.
        """

        return self._classes

    def __iter__(self):
        """
        Allows iteration through our known class data
        in async form.
        """

        for class_name, class_data in list(self._classes.items()):
            yield (class_name, class_data)

    def __import_module(self, module_name: str) -> object:
        """
        Imports the requested module by its module
        name import path
        """

        assert module_name != ''
        assert module_name != None

        components = module_name.split('.')
        module_path = '.'.join(components[:-1])
        module = importlib.import_module(module_path)

        return module
    
    def is_registered(self, class_name: str) -> bool:
        """
        Returns true if the class name is already registered
        with the class registry singleton
        """

        return class_name in self._classes

    def batch_register_classes(self, class_list: list, meta_list: list = []) -> None:
        """
        Batch registers classes for module setups
        """

        self.notify.debug('Batch registering %df classes...' % len(class_list))
        for class_info in class_list:
            class_name, module_name, meta = class_info
            self.notify.debug('Registering %s from module (%s) with meta (%s)' % (
                class_name, module_name, str(meta)))
            self.register_class(class_name, module_name, **meta)

        self.notify.debug('Setting %d meta key/values' % (len(meta_list)))
        for meta_info in meta_list:
            class_name, meta_key, meta_value = meta_info
            self.set_class_meta(class_name, meta_key, meta_value)

    def register_class(self, class_name: str, module_name: str, **meta) -> None:
        """
        Registers a new class with the class registry singleton
        """    
        return self.register_class_alias(class_name, class_name, module_name, **meta)    

    def register_class_alias(self, class_alias: str, class_name: str, module_name: str, **meta) -> None:
        """
        Registers a new class using an alias with the class registry singleton
        """

        # Verify the class does not already exist
        if self.is_registered(class_name):
            self.notify.warning('Failed to register class. Class "%s" already exists in the registry' % (
                class_alias))
            
            return

        self.notify.debug('Registering class (%s) as "%s"' % (class_name, class_alias))
        self._classes[class_alias] = (class_name, module_name, None, meta)

    def unregister_class(self, class_name: str) -> None:
        """
        Attempts to unregister a class from the class registry singleton
        """

        # Verify the class exists
        if not self.is_registered(class_name):
            self.notify.warning('Failed to unregister class. Class "%s" is not registered with the singleton!' % (
                class_name))

            return

        # Remove the class from the registry
        self._classes.pop(class_name)

    def get_class(self, class_name: str) -> object:
        """
        Attempts to retrieve a class from the class registry
        if it exists
        """

        # Verify the class exists
        if not self.is_registered(class_name):
            self.notify.warning('Class "%s" is not registered with the %s singleton!' % (
                class_name, self.__class__.__name__))

            return

        # Retrieve our known information about the class
        cls_name, module_name, module, meta = self._classes[class_name]

        # Update the registry if a module instance does not already exist
        if not module:
            module = self.__import_module(module_name)
            self._classes[class_name] = (cls_name, module_name, module, meta)

        # Attempt to retrieve the class instance and return it to
        # the requester
        cls = getattr(module, cls_name, None)
        if not cls:
            self.notify.warning('Failed to retrieve class. Class unable to be imported: %s' % (
                class_name))

            return

        return cls

    def get_class_meta(self, class_name: str, meta_tag: str = None, default: object = None)-> object:
        """
        Attempts to retrieve the requested meta value from 
        a registered class 
        """

        # Verify the class exists
        if not self.is_registered(class_name):
            self.notify.warning('Failed to get meta. Class "%s" is not registered with the singleton!' % (
                class_name))

            return

        # Retrieve the requested tag from our known class
        # in the registry
        cls_name, module_name, module, meta = self._classes[class_name]
#
        if meta_tag is None:
            return meta
        else:
            return meta.get(meta_tag, default)

    def set_class_meta(self, class_name: str, meta_tag: str, meta_value: object) -> None:
        """
        Attempts to set a meta value on a class within the class registry
        """

        # Verify the class exists
        if not self.is_registered(class_name):
            self.notify.warning('Failed to set meta. Class "%s" is not registered with the singleton!' % (
                class_name))

            return

        # Set the meta key/value for the requested class
        cls_name, module_name, module, meta = self._classes[class_name]
        self.notify.debug('Setting class meta for %s. %s = %s' % (
            cls_name, meta_tag, meta_value))
        meta[meta_tag] = meta_value
    
    def query_meta(self, **meta) -> list:
        """
        Queries the class registry for all class meta data 
        matching the requested query
        """

        classes = []

        # Iterate over all known classes 
        # to compare meta data. Returning all known
        # classes that match the query data
        for class_name in self._classes:
            cls_name, module_name, module, meta = self._classes[class_name]

            for meta_tag, meta_value in list(meta.items()):
                if meta_tag in meta and meta[meta_tag] == meta_value:
                    cls = self.get_class(class_name)
                    if cls:
                        classes.append(cls)

        return classes

class ClassRegistryMixIn(object):
    """
    """

    def _get_registry(self) -> ClassRegistry:
        """
        Retrieves the current class registry
        singleton
        """

        return ClassRegistry.get_singleton()

    def _get_class(self, *args, **kwargs) -> object:
        """
        """

        return self._get_registry().get_class(*args, **kwargs)

    def _get_class_meta(self, *args, **kwargs) -> object:
        """
        """

        return self._get_registry().get_class_meta(*args, **kwargs)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from direct.showbase.DirectObject import DirectObject

from panda3d_gemstone.logging import utilities as logging
from panda3d_gemstone.framework.utilities import get_snake_case

class InternalObject(object):
    """
    Base class for all Gemstone framework internal objects
    """

    def __init__(self, *args, **kwargs):
        self.notify = logging.get_notify_category(self.get_notify_name())

    def get_notify_name(self) -> str:
        """
        Returns this object's notifier name
        """

        return get_snake_case(self.__class__.__name__, splitter='-')

    def destroy(self) -> None:
        """
        Called on object destruction
        """

    def __str__(self) -> str:
        """
        Custom base string handler for debugging 
        internal gemstone objects
        """

        return self.__class__.__name__

class PandaBaseObject(DirectObject):
    """
    Legacy base class of all Gemstone framework internal objects that
    utilize Panda3d internal code
    """

    def __init__(self, *args, **kwargs):
        DirectObject.__init__(self, *args, **kwargs)
        self.__notify = logging.get_notify_category(self.get_notify_name())
        self.notify.warning('%s is inheriting from legacy object: PandaBaseObject' % (
            self.__class__.__name__))

    @property
    def notify(self) -> object:
        """
        Returns this object's notifier logger
        """

        return self.__notify

    def get_notify_name(self) -> str:
        """
        Returns this object's notifier name
        """

        return get_snake_case(self.__class__.__name__, splitter='-')

    def destroy(self) -> None:
        """
        Called on object destruction
        """

        self.ignore_all()

    def __str__(self) -> str:
        """
        Custom base string handler for debugging 
        internal gemstone objects
        """

        return self.__class__.__name__


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d.core import CollisionTraverser, CollisionHandlerFluidPusher, CollisionHandlerQueue
from panda3d.core import CollisionNode, BitMask32, NodePath

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.runnable import Runnable

from panda3d_gemstone.engine import runtime

class CollisionSystem(Singleton, Runnable):
    """
    Manages all native Panda3D collision interactions inside
    the Gemstone framework
    """

    def __init__(self):
        Singleton.__init__(self)
        Runnable.__init__(self, priority=25)

        self.__traversers = {}
        self.__collision_handlers = {}
        self.__collision_objects = {}
        self.add_traverser('default')
        self.add_handler('queue', CollisionHandlerQueue())
        self.add_handler('pusher', CollisionHandlerFluidPusher())
        self.__debugging = False
    
    def destroy(self) -> None:
        """
        Destroys the collision system singleton instance
        """

        self.__traversers = {}
        self.__collision_handlers = {}
        self.__collision_objects = {}
        Runnable.deactive(self)

    def has_collision_object(self, name: str) -> bool:
        """
        Returns true if the collision object exists in the system
        """

        return name in self.__collision_objects

    def get_collision_object(self, name: str) -> object:
        """
        Returns the collision object if it exists. Otherwise
        returns NoneType
        """

        return self.__collision_objects.get(name, None)

    def show_collision_object(self, name: str) -> None:
        """
        Shows the collision object's NodePath
        """

        if not self.has_collision_object(name):
            return

        nodepath = self.get_collision_object(name)
        nodepath.show()

    def hide_collision_object(self, name: str) -> None:
        """
        Hides the collision object's NodePath
        """

        if not self.has_collision_object(name):
            return

        nodepath = self.get_collision_object(name)
        nodepath.hide()

    def hide_all_collision_objects(self) -> None:
        """
        Hides all collision objects in the system
        """

        for name in self.__collision_objects:
            self.hide_collision_object(name)

    def show_all_collision_objects(self) -> None:
        """
        Shows all collision objects in the system
        """

        for name in self.__collision_objects:
            self.show_collision_object(name)

    def add_traverser(self, name: str, priority: int = 0, traverser_class: object = CollisionTraverser) -> None:
        """
        Adds a new collision traverser to the collision system
        """

        if self.has_traverser(name):
            self.notify.warning('Failed to add traverser. Traverser (%s) already exists!' % name)
            return

        traverser = traverser_class()
        self.__traversers[name] = (priority, traverser)
        traverser.set_respect_prev_transform(True)

    def has_traverser(self, name: str) -> bool:
        """
        Returns true if the collision traverser exists
        """

        return name in self.__traversers

    def get_traverser(self, name: str) -> object:
        """
        Returns the collision traverser if it exists. Otherwise
        returns NoneType
        """

        return self.__traversers.get(name, None)

    def remove_traverser(self, name: str) -> None:
        """
        Removes a traverser from the collision system by its name
        """

        if not self.has_traverser(name):
            self.notify.warning('Failed to remove traverser. %s does not exist' % name)
            return

        if name == 'default':
            self.notify.warning('Failed to remove traverser. Cannot delete default')
            return

        del self.__traversers[name]

    def show_collisions(self) -> None:
        """
        Enables the debugging collision visualization and shows all collisions
        """

        self.__debugging = False
        for priority, traverser in lisT(self.__traversers.values()):
            if hasattr(traverser, 'show_collisions'):
                traverser.show_collisions(runtime.render)

    def hide_collision(self) -> None:
        """
        Disables the debugging collision visualization and hides all collisions
        """

        self.__debugging = False
        for priority, traverser in lisT(self.__traversers.values()):
            if hasattr(traverser, 'hide_collisions'):
                traverser.hide_collisions(runtime.render)

    def toggle_collisions(self) -> None:
        """
        Toggles the collision debugging visuals
        """

        if self.__debugging:
            self.hide_collision()
        else:
            self.show_collisions()

    def has_handler(self, name: str) -> bool:
        """
        Returns true if the collision handler exists
        """

        return name in self.__collision_handlers

    def add_handler(self, name: str, handler: object) -> object:
        """
        Adds a new collision handler if it does not already exist
        """

        if self.has_handler(name):
            self.notify.warning('Failed to add new handler. Handler (%s) already exists' % name)
            return None

        self.__collision_handlers[name] = handler
        return handler

    def remove_handler(self, name: str) -> None:
        """
        Attempts to remove a handler if it exists
        """

        if not self.get_handler(name):
            self.notify.warning('Failed to remove handler %s. Handler does not exist' % name)
            return

        del self.__collision_handlers[name]

    def get_handler(self, name: str) -> object:
        """
        Returns the handler object if it exists. Otherwise
        returns NoneType
        """

        return self.__collision_handlers.get(name, None)

    def add_collision_object(self, name: str, parent: object, obj: object) -> object:
        """
        Adds a collision object to the system
        """

        if self.has_collision_object(name):
            self.notify.warning('Attempted to add collision object that already exists (%s)!' % name)
            return self.get_collision_object(name)

        if type(obj) == type([]):
            objs = obj
        else:
            objs = [obj]

        collision_node = CollisionNode(name)
        for o in objs:
            collision_node.add_solid(o)

        collision_node.set_into_collide_mask(BitMask32.all_off())
        if parent:
            collision_path = parent.attach_new_node(collision_node)
        else:
            collision_path = NodePath(collision_node)

        self.__collision_objects[name] = collision_path
        return collision_path

    def enable_collision_object(self, object_name: str, handler_name: str, traverser_name: str = 'default') -> None:
        """
        Enables a collision object in the system
        """

        if not self.has_collision_object(object_name):
            self.notify.warning('Failed to remove collision object. Object "%s" does not exist' % object_name)
            return
        
        if not self.has_handler(handler_name):
            self.notify.warning('Failed to remove collision object. Handler "%s" does not exist' % handler_name)
            return      

        if not self.has_traverser(traverser_name):
            self.notify.warning('Failed to remove collision object. Traverser "%s" does not exist' % traverser_name)
            return

        nodepath = self.get_collision_object(object_name)
        handler = self.get_handler(handler_name)
        priority, traverser = self.get_traverser(traverser_name)
        traverser.add_collider(nodepath, handler)

        if isinstance(handler, CollisionHandlerFluidPusher) and runtime.base.drive:
            handler.add_collider(nodepath, nodepath.get_parent(), runtime.base.drive.node())

    def remove_collision_object(self, object_name: str, handler_name: str, traverser_name: str = 'default') -> None:
        """
        Removes a collision object from the system
        """

        if not self.has_collision_object(object_name):
            self.notify.warning('Failed to remove collision object. Object "%s" does not exist' % object_name)
            return
        
        if not self.has_handler(handler_name):
            self.notify.warning('Failed to remove collision object. Handler "%s" does not exist' % handler_name)
            return      

        if not self.has_traverser(traverser_name):
            self.notify.warning('Failed to remove collision object. Traverser "%s" does not exist' % traverser_name)
            return

        nodepath = self.get_collision_object(object_name)
        handler = self.get_handler(handler_name)
        priority, traverser = self.get_traverser(traverser_name)
        traverser.remove_collider(nodepath)

        if isinstance(handler, CollisionHandlerFluidPusher):
            handler.remove_collider(nodepath)

        del self.__collision_objects[object_name]

    async def tick(self, dt: float) -> None:
        """
        Called once per frame to perform collision traversal
        """

        traversers = list(self.__traversers.values())
        traversers.sort()

        for priority, traversers in traversers:
            traversers.traverse(runtime.render)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import fnmatch
import logging
from hashlib import md5

import direct.stdpy.file
import glob

from panda3d.core import VirtualFileSystem, Filename, Multifile
from panda3d.core import VirtualFileMountHTTP, VirtualFileMountRamdisk
from panda3d.core import VirtualFileMountSystem, VirtualFileMountMultifile
from panda3d.core import get_model_path

from panda3d_gemstone.logging import utilities as logging

__file_notify = logging.get_notify_category('file-system')

def os_path_exists(path: str) -> bool:
    """
    OS function for verifying if a file exists
    """

    return os.path.exists(path)

def os_path_is_dir(path: str) -> bool:
    """
    OS function for checking if a path is a 
    directory
    """

    return os.path.isdir(path)

def os_get_file_date(path: str) -> int:
    """
    OS function for retrieving the file
    date from a path
    """

    exists = os.path.exists(path)
    date = 0 

    if exists:
        stat = os.stat(path)
        date = stat.st_mtime

    return date

def os_get_file_size(filehandle: object) -> float:
    """
    OS function for retrieving the file
    size
    """

    stat = os.fstat(filehandle.fileno())
    return stat.st_size

def os_get_matching_files(path: str, pattern: str) -> list:
    """
    OS function for retrieving all
    pattern matching files in the directory
    """

    return glob.glob(os.path.join(path, pattern))

_vfs = VirtualFileSystem.get_global_ptr()

def vfs_path_exists(path: str) -> bool:
    """
    Panda3D VFS function for verifying if a path exists
    """

    return _vfs.exists(Filename(path))

def vfs_path_is_dir(path: str) -> bool:
    """
    Panda3D VFS function for verifying if a path is a 
    directory
    """

    return vfs.is_directory(path)

def vfs_get_file_date(path: str) -> int:
    """
    Panda3d VFS function for retrieving a file 
    date
    """

    fh = _vfs.get_file(Filename(path), status_only=True)
    date = 0

    if fh:
        date = fh.get_timestamp()

    return date

def vfs_get_file_size(filehandle: object) -> float:
    """
    Panda3D VFS function for retrieving a files
    size
    """

    fh = _vfs.get_file(Filename(filehandle.name), status_only=True)
    if fh == None:
        __file_notify.warning('Failed to retrieve file size for file: %s. Defaulting to -1' % filehandle.name)
        return -1

    return fh.get_file_size(filehandle._file__stream)

def vfs_get_matching_files(path: str, pattern: str) -> list:
    """
    Panda3D VFS function for retrieving all
    files in the directory matching
    the string pattern
    """

    found_files = []
    result = _vfs.scan_directory(Filename(path))
    if result:
        for i in range(result.get_num_files()):
            filename = result.get_file(i).get_filename().get_basename()
            if fnmatch.fnmatch(filename, pattern):
                found_files.append(fixed_join(path, filename))

    return found_files

def fix_path(path: str) -> str:
    """
    Replaces the system path sep with a forward slash
    """

    return path.replace(os.sep, '/')

def fixed_join(lhs: str, rhs: str) -> str:
    """
    Joins two paths together and then runs fix_path on the resulting
    path string
    """

    return fix_path(os.path.join(lhs, rhs))

def get_file_extension(path: str) -> str:
    """
    Returns the path's file extension if present
    """

    parts = path.split('.')
    if len(parts) > 1:
        return parts[-1]
    return None

def file_md5(path: str, blocksize: int = 1024) -> object:
    """
    Returns a files md5 hash from its path
    """

    m = md5()
    if path_exists(path):
        fh = file(path, 'rb')
        data = [1]
        accumulated = 0

        while len(data) > 0:
            data = fh.read(blocksize)
            accumulated += len(data)
            m.update(data)

        fh.close()
    else:
        __file_notify.warning('Failed to compute MD5 has for %s. File does not exist' % path)

    return m.digest()

def correct_path_case(path: str) -> str:
    """
    """

    if path and os_path_exists(path):
        path = fix_path(path)
        path_components = path.split('/')
        
        current_path = ''
        fixed_path = ''

        for pc in path_components:
            current_path = os.path.join(current_path, pc)
            dirpath = os.path.split(os.path.realpath(current_path))[0]
            if os_path_exists(dirpath):
                files = os.listdir(dirpath)
                if pc in files or pc == '.' or pc == '..':
                    fixed_path = os.path.join(fixed_path, pc)
                else:
                    files_lower = [filename.lower() for filename in files]
                    pc_lower = pc.lower()

                    if pc_lower in files_lower:
                        idx = files_lower.index(pc_lower)
                        fixed_path = os.path.join(fixed_path, files[idx])
            else:
                return path

        return fix_path(fixed_path)

    return path

def check_file_path(path: str) -> str:
    """
    """

    if not path:
        return

    return correct_path_case(path)

__path_exists_func = os_path_exists
__path_is_dir_func = os_path_is_dir
__get_file_date_func = os_get_file_date
__get_file_size_func = os_get_file_size
__get_matching_files_func = os_get_matching_files

def path_exists(path: str) -> bool:
    """
    Returns true if the requested path exists in the applications
    file system
    """

    global __path_exists_func
    return __path_exists_func(path)

def path_is_dir(path: str) -> bool:
    """
    Returns true if the requested path is a directory
    in the applications file system
    """

    global __path_exists_func
    return __path_exists_func(path)

def get_file_date(path: str) -> int:
    """
    Returns the date information from
    a file in the applications file system
    """

    global __get_file_date_func
    return __get_file_date_func(path)

def get_file_size(path: str) -> float:
    """
    Returns the size information from
    a file in the applications file system
    """

    global __get_file_size_func
    return __get_file_size_func(path)

def get_matching_files(path: str, pattern: str) -> list:
    """
    Returns all files in the directory
    matching the requested pattern
    """

    global __get_matching_files_func
    return __get_matching_files_func(path, pattern)

def __switch_configurable_cache(vfs: bool = False) -> None:
    """
    Switches the configurable object's cache system to and from
    the Panda3D virtual file system and the operation systems
    native io functions
    """

    from panda3d_gemstone.framework.configurable import Configurable
    from panda3d_gemstone.framework.configurable import OSConfigurableCache
    from panda3d_gemstone.framework.configurable import VFSConfigurableCache

    if vfs:
        Configurable.__Configurable__cache = VFSConfigurableCache()
    else:
        Configurable.__Configurable__cache = OSConfigurableCache()

def switch_file_functions_to_vfs() -> None:
    """
    Switches the file access functions to use the Panda3D
    Virtual File System
    """

    global __path_exists_func
    global __path_is_dir_func
    global __get_file_size_func
    global __get_file_date_func
    global __get_matching_files_func

    __path_exists_func = vfs_path_exists
    __path_is_dir_func = vfs_path_is_dir
    __get_file_size_func = vfs_get_file_size
    __get_file_date_func = vfs_get_file_date
    __get_matching_files_func = vfs_get_matching_files
    __switch_configurable_cache(True)

def switch_file_functions_to_os() -> None:
    """
    Switches the file access functions to use the Operating
    System's IO functions
    """

    global __path_exists_func
    global __path_is_dir_func
    global __get_file_size_func
    global __get_file_date_func
    global __get_matching_files_func

    __path_exists_func = os_path_exists
    __path_is_dir_func = os_path_is_dir
    __get_file_size_func = os_get_file_size
    __get_file_date_func = os_get_file_date
    __get_matching_files_func = os_get_matching_files
    __switch_configurable_cache(False)

def vfs_get_mount_count() -> int:
    """
    Returns the number of individual mounts in the Virtual
    File System
    """

    return _vfs.get_num_mounts()

def vfs_get_mount(index: int) -> object:
    """
    Returns the nth mount in the Virtual File System
    """

    return _vfs.get_mount(index)

def vfs_unmount_all_directories() -> None:
    """
    Unmounts all directories from the virtual file system
    """

    _vfs.unmount_all()

def vfs_mount_directory(mount_point: str, directory: str) -> bool:
    """
    Mounts the directory to the requested mount point.
    Returning the result of the mounting operation
    """

    result = _vfs.mount(
        Filename(directory),
        Filename(mount_point),
        VirtualFileSystem.MF_read_only)

    if not result:
        __file_notify.warning('Failed to mount directory (%s) to (%s)!' % (
            directory, mount_point))

    switch_file_functions_to_vfs()
    return result

def vfs_mount_url(mount_ptr: str, url: str) -> bool:
    """
    Mounts a http directory to the requested mount point.
    Returning the result of the mounting operation. Best used with
    CDN or download like servers
    """

    raise NotImplementedError('vfs_mount_url is currently not supported')

    if not result:
        __file_notify.warning('Failed to mount directory (%s) to (%s)!' % (
            directory, mount_point))

    switch_file_functions_to_vfs()
    return result

def vfs_mount_multifile(mount_point: str, multifile: str) -> bool:
    """
    Mounts the multifile to the requested mount point
    if it exists. Returning the result of the mounting
    operation
    """

    if not os_path_exists(multifile):
        return False

    __file_notify.info('Mounting MF: %s' % vfs_mount_multifile)
    
    mf = Multifile()
    mf.open_read(multifile)

    result = vfs.mount(mf, Filename(mount_point))

    if not result:
        __file_notify.warning('Failed to mount multifile (%s) to (%s)!' % (
            multifile, mount_point))

    switch_file_functions_to_vfs()
    return result

def vfs_mount_subdirectories(mount_point: str, root_dir: str, mount_root: bool = True) -> bool:
    """
    Walks through a root directory and mounts all subdirectorys
    to the requested mount point. Generally used for development
    source version of an application or servers.
    """

    folders = [os.path.join(root_dir, o) for o in os.listdir(root_dir) if os.path.isdir(os.path.join(root_dir,o))]
    if mount_root:
        if not root_dir.endswith('/'):
            root_dir += '/'
        folders.append(root_dir)

    for folder in folders:
        
        # Ignore invalid folders
        base_name = os.path.basename(folder)
        if base_name.startswith('.'):
            continue

        folder = Filename.from_os_specific(os.path.abspath(folder))
        __file_notify.debug('Mounting (%s) to (%s)' % (folder, mount_point))
        vfs_mount_directory(mount_point, folder)

def append_subdirectories_to_search(root_dir: str, mount_root: bool = True) -> None:
    """
    Walks through a root directory and mounts all subdirectorys
    to the search path. Generally used for development
    source version of an application or servers.
    """

    model_path = get_model_path()
    folders = [os.path.join(root_dir, o) for o in os.listdir(root_dir) if os.path.isdir(os.path.join(root_dir,o))]
    if mount_root:
        if not root_dir.endswith('/'):
            root_dir += '/'
        folders.append(root_dir)

    for folder in folders:

        # Ignore invalid folders
        base_name = os.path.basename(folder)
        if base_name.startswith('.'):
            continue

        folder = Filename.from_os_specific(os.path.abspath(folder))
        __file_notify.debug('Mounting (%s) to search path' % (folder)) 
        model_path.append_directory(folder)

def get_search_path() -> object:
    """
    Returns the panda3d model search path
    """

    return get_model_path()

def reload_vfs_mount_urls() -> None:
    """
    Reads all of the vfs-mount-url lines in the Config.prc file and replaces the mount settings to match them.
    This will mount any url’s mentioned in the config file, and unmount and unmount any url’s no longer mentioned in the config file. 
    Normally, it is called automatically at startup, and need not be called again, unless you have fiddled with some config settings.
    """

    VirtualFileMountHTTP.reload_vfs_mount_url()

def is_path_multifile(path: str) -> bool:
    """
    Returns true if the requested path is a multifile
    """

    dir_exists = path_exists(path)
    if dir_exists:
        return False

    return path_exists('%s.mf' % path)

def vfs_enable_python_imports() -> None:
    """
    Enables VFS based Python import access for the Gemstone
    application
    """

    from direct.showbase.VFSImporter import register
    register()

def vfs_refresh_python_packages() -> None:
    """
    Refreshes the Python packages in the virtual file system
    """

    from direct.showbase.VFSImporter import reloadSharedPackages
    reloadSharedPackages()

def vfs_refresh_python_package(mod: object) -> None:
    """
    Refreshes the requested Python package in the virtual file system
    """

    from direct.showbase.VFSImporter import reloadSharedPackage
    reloadSharedPackage(mod)


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json

from panda3d_gemstone.io.file_system import get_file_extension, path_exists

from panda3d_gemstone.framework.exceptions import raise_not_implemented
from panda3d_gemstone.framework.internal_object import InternalObject

def load_file(file_path: str) -> object:
    """
    Attempts to load a dictionary file from the application's
    file system
    """

    dict_file = DictionaryFile()
    success = dict_file.load(file_path)
    if not success:
        return None

    return dict_file

class DictionaryFile(InternalObject):
    """
    Represents a dictionary file in the Gemstone framework
    """

    def __init__(self):
        InternalObject.__init__(self)
        DataContainer.__init__(self)

        self.__data = {}

    @property
    def data(self) -> dict:
        """
        Data contained inside the dictionary file
        """

        return self.get_data()

    def get_data(self) -> dict:
        """
        Returns the data loaded from the dictionary file
        """

        return self.__data

    def load(self, file_path: str) -> bool:
        """
        Loads a dictionary file from the application's file system
        """

        if not path_exists(file_path):
            self.notify.wanring('Attempted to load nonexistance file: %s' % file_path)
            return False

        ext = get_file_extension(file_path).lower()
        ext_handler = getattr(self, '__parse_%s' % ext)
        if not ext_handler:
            self.notify.warning('Attempted to load unsupported file type: %s' % ext)
            return False

        data = open(file_path, 'r')
        try:
            ext_handler(data)
        except Exception as e:
            self.notify.warning('Failed to parse dictionary file: %s' % file_path)
            return False

        return True

    def write(self, file_path: str) -> None:
        """
        Writes a dictionary file to the application's file system
        """

        raise_not_implemented(self, 'write')

    def __parse_json(self, data: str) -> None:
        """
        Parses incoming json data
        """

        incoming = json.loads(data)
        for key, value in incoming.items():
            self.__data[key] = value
        




"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import logging

from panda3d_gemstone.engine import runtime
from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.runnable import Runnable

from panda3d.bullet import BulletWorld, BulletDebugNode
from panda3d.core import Vec3

class PhysicsEngine(Singleton, Runnable):
    """
    Singleton representing bullet physics in the panda3d game 
    engine.
    """

    def __init__(self):
        Runnable.__init__(self)
        Singleton.__init__(self)

        self.__world = None
        self.__debug_node = None
        self.__debug_node_np = None
        self.setup()

    @property
    def debug_mode(self) -> bool:
        """
        Returns true when the physics engine has the debug node
        enabled
        """

        return self.__debug_node != None

    @property
    def physics_world(self) -> BulletWorld:
        """
        Returns this physics engines bullet world instance
        """

        return self.__world

    @property
    def gravity(self) -> Vec3:
        """
        Returns this physics engines current gravity configuration
        as a property
        """

        return self.get_gravity()

    @gravity.setter
    def gravity(self, gravity: Vec3) -> None:
        """
        Sets this physics engines current gravity as a setter
        """

        self.set_gravity(gravity)

    def setup(self) -> None:
        """
        Performs setup operations on the singleton
        """

        self.__world = BulletWorld()
        self.set_gravity(Vec3(0, 0, -9.81))
        self.activate()

    def destroy(self) -> None:
        """
        Performs shutdown operations on the singleton
        """

        # Remove debug node if present
        if self.__debug_node:
            self.__debug_node.remove_node()

        # Deactivate runnable
        self.deactivate()

    def toggle_debug(self) -> None:
        """
        Toggles the current debug state
        """

        assert self.__world != None
        if self.debug_mode:
            self.__world.set_debug_node(None)
            self.__debug_node_np.remove_node()
        else:
            self.__debug_node = BulletDebugNode('%s-Debug' % self.__class__.__name__)
            self.__debug_node_np = runtime.render.attach_new_node(self.__debug_node)
            self.__debug_node_np.show()
            self.__world.set_debug_node(self.__debug_node)

    async def tick(self, dt: float) -> None:
        """
        Called once per frame to perform physics world updates
        """

        if not self.__world:
            return

        self.__world.do_physics(dt, 10, 1.0/180.0)

    def __getattr__(self, key: str) -> object:
        """
        Custom attribute getter to pass through unknown attributes
        to the physics world instance
        """

        result = None
        if self.__world != None:
            if hasattr(self.__world, key):
                result = getattr(self.__world, key)
        
        if not result:
            raise AttributeError('%s instance does not have attribute %s' % (
                self.__class__.__name__, key))

        return result


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d_gemstone.framework.singleton import Singleton
from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.framework.utilities import get_snake_case

from panda3d_gemstone.engine.shader import load_shader
from panda3d_gemstone.engine import runtime

from panda3d.core import NodePath

class GraphicsManager(Singleton, Configurable, InternalObject):
    """
    """

    def __init__(self, config_path: str):
        Singleton.__init__(self)
        InternalObject.__init__(self)
        self.__shader_data = {}
        self.__shader_defines = {}
        Configurable.__init__(self, config_path)
        self.__activated = False
        self.initialize()

    def load_shader_data(self, data: dict) -> None:
        """
        Loads the shader data into memory
        """

        self.__shader_data = data

    def load_defines_data(self, data: dict) -> None:
        """
        Loads the graphics define data into memory
        """

        for key, value in data.items():
            define_name = get_snake_case(key).upper()
            self.__shader_defines[define_name] = value

    def is_active(self) -> bool:
        """
        Returns true if the graphics manager is active
        """

        return self.__activated

    def activate(self, root_node: NodePath) -> None:
        """
        Activates the graphics manager on the requested root node 
        instance
        """

        if self.__activated:
            self.notify.warning('Failed to activate %s. Already active' % (
                self.__class__.__name__))
        
            return

        assert root_node != None
        assert runtime.has_base()
        assert isinstance(root_node, NodePath)

        # Setup shaders
        core_shader = load_shader(
            vertex=self.__shader_data.get('coreVertexPath', None),
            fragment=self.__shader_data('coreFragmentPath', None))
        assert core_shader != None
        root_node.set_shader(core_shader)

        # Set general inputs
        root_node.set_shader_input('camera', runtime.base.cam)



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from math import sqrt

from panda3d.core import NodePath, AmbientLight, DirectionalLight
from panda3d.core import PointLight, Spotlight, PerspectiveLens
from panda3d.core import Vec3, Vec4
from panda3d.core import PointLight as PandaPointLight
from panda3d.core import BoundingSphere

from panda3d_gemstone.framework.configurable import Configurable
from panda3d_gemstone.framework.utilities import delegate, perform_dependency_check
from panda3d_gemstone.framework.utilities import perform_child_class_check

from panda3d_gemstone.world.entity import Entity
from panda3d_gemstone.engine import runtime

PandaLights = {
    'point': PointLight,
    'ambient': AmbientLight,
    'directional': DirectionalLight,
    'spot': Spotlight
}

class Light(Configurable, NodePath):
    """
    Base class for all Gemstone lighting objects
    """

    @staticmethod
    def insert(light: object, parent: object) -> None:
        """
        Inserts the light object into the scene graph
        """

        light.reparent_to(parent)
        parent.set_light(light)

    @staticmethod
    def remove(light: object) -> None:
        """
        Removes the light from the scene graph
        """

        if not light.get_parent().is_empty():
            light.get_parent().clear_light(light)

        light.remove_node()
 
    def __init__(self, config_path: str, section: str = 'Configuration'):
        NodePath.__init__(self, '')
        Configurable.__init__(self, config_path, section)
        self.__light_type = self.pop('type')

    @property
    def light_type(self) -> str:
        """
        Returns the light's engine type as a property
        """

        return self.get_light_type()

    def get_light_type(self) -> str:
        """
        Returns this light's engine type
        """

        return self.__light_type

class PandaLight(Light):
    """
    Represents a light in the Gemstone framework. Acts
    as a wrapper object around the Panda3D lighting
    system
    """

    def __init__(self, config_path: str, section: str = 'Configuration'):
        super().__init__(config_path, section)
        self.__light = self.__create_light_instance()

        if self.__type == 'spot':
            self.__light.set_len(PerspectiveLens())
            self.assign(NodePath(self.__light.upcast_to_LensNode()))
        else:
            self.assign(NodePath(self.__light.upcast_to_PandaNode()))

        self.initialize()

    @property
    def light(self) -> object:
        """
        Returns the light's engine light object
        as a property
        """

        return self.get_light()

    def get_light(self) -> object:
        """
        Returns the native Panda3D engine light object
        for this gemstone light
        """

        return self.__light

    def __create_light_instance(self) -> object:
        """
        Creates a new light instance from the panda
        light's configured type
        """

        light_type = self.get_light_type()
        panda_type = PandaLights.get(light_type, None)
        assert panda_type != None

        return panda_type(light_type + 'Light')

    def set_pos(self, *pos) -> None:
        """
        Sets the lighting object's position
        """

        NodePath.set_pos(self, *pos)
        self.set_point(*pos)

    def __getattr__(self, name) -> object:
        """
        Custom getattr to pass through to the child light 
        object
        """

        def fail() -> None:
            """
            Throws an AttributeError exception
            """

            raise AttributeError('Unknown attribute: %s' % name)

        # Verify we have a light
        if self.__light is None:
            fail()

        if hasattr(self.__light, name):
            return getattr(self.__light, name)
        else:
            fail()

class RenderPipelineLight(Light):
    """
    Represents a light in the Gemstone framework. Acts
    as a wrapper object around the Panda3D Render Pipeline
    lighting system
    """

    def __init__(self, config_path: str, section: str = 'Configuration'):
        super().__init__(config_path, section)
        perform_dependency_check('rpcore')

        raise NotImplementedError('%s is not yet implemented!' % self.__class__.__name__)

class AmbientLightMixIn(object):
    """
    """

    def calc_light_color(self, point, normal) -> object:
        """
        """
        
        perform_child_class_check('AmbientLightMixIn', self.__class__, 'get_color')
        return self.get_color()

class DirectionalLightMixIn(object):
    """
    """

    def calc_light_color(self, point, normal) -> object:
        """
        """

        perform_child_class_check('DirectionalLightMixIn', self.__class__, 'get_direction')
        perform_child_class_check('DirectionalLightMixIn', self.__class__, 'get_color')

        direction = self.get_direction()
        intensity = -Vec3(direction[0], -direction[2], direction[1]).dot(normal)

        if intensity <= 0:
            intensity = 0

        return self.get_color * intensity

class PointLightMixIn(object):
    """
    """

    def calc_bounding_sphere(self) -> object:
        """
        """
        
        perform_child_class_check('PointLightMixIn', self.__class__, 'get_point')

        point = self.get_point()
        radius = self.calc_relevant_radius()

        if radius > 0:
            return BoundingSphere(center=point, radius=radius)
        
        return None

    def calc_relevant_radius(self) -> float:
        """
        """

        perform_child_class_check('PointLightMixIn', self.__class__, 'get_attenuation')

        max_att = 32.0
        attenuation_vec = self.get_attenuation()
        if attenuation_vec[2] != 0:

            sqr = attenuation_vec[1] * attenuation_vec[1] - 4 * (attenuation_vec[2] * attenuation_vec[0] - max_att)
            if sqr < 0:
                return 0

            sqr_val = sqrt(sqr)
            r1 = (-attenuation_vec[1] + sqr_val) / (2 * attenuation_vec[2])
            r2 = (-attenuation_vec[1] - sqr_val) / (2 * attenuation_vec[2])

            if r1 > r2:
                return r1
            else:
                return r2
        elif attenuation_vec[1] != 0:
            return (max_att - attenuation_vec[0]) / attenuation_vec[1]
        else:
            return 0

    def calc_light_intensity(self, point: object) -> float:
        """
        """

        perform_child_class_check('PointLightMixIn', self.__class__, 'get_point')
        perform_child_class_check('PointLightMixIn', self.__class__, 'get_attenuation')

        point_to_light_vec = self.get_point() - point
        distance_vec = Vec3(1, point_to_light_vec.length(), point_to_light_vec.length_squared())
        intensity = self.get_attenuation().dot(distance_vec)

        if intensity > 0:
            intensity = 1.0 / intensity
        else:
            intensity = 0

        return intensity

    def calc_light_color(self, point, normal) -> object:
        """
        """

        perform_child_class_check('PointLightMixIn', self.__class__, 'get_point')
        perform_child_class_check('PointLightMixIn', self.__class__, 'get_color')

        point_to_light_vec = self.get_point() - point
        point_to_light_vec.normalize()

        intensity = point_to_light_vec.dot(normal)
        if intensity <= 0:
            intensity = 0
        else:
            intensity *= self.calc_light_intensity(point)

        return self.get_color() * intensity

class LightSet(Entity):
    """
    """

    class GemDirectionalLight(DirectionalLight, DirectionalLightMixIn):
        """
        """

        def __init__(self, name: str):
            DirectionalLight.__init__(self, name)
            self.set_color(Vec4(0, 0, 0, 1))

    class GemAmbientLight(AmbientLight, AmbientLightMixIn):
        """
        """

        def __init__(self, name: str):
            AmbientLight.__init__(self, name)
            self.set_color(Vec4(0, 0, 0, 1))

    def __init__(self, config_path: str):
        self.__ambient = LightSet.GemAmbientLight('ambient')
        self.__directional_lights = [LightSet.GemDirectionalLight('directional%d' % (i + 1)) for i in range(2)]
        self.__light_nps = []
        super().__init__(config_path)

    def destroy(self) -> None:
        """
        Destroys the light set entity
        """

        self.deactivate()
        Entity.destory(self)

    def get_relevant_lights(self) -> list:
        """
        """

        return [light for light in [self.__ambient] + self.__directional_lights if light.get_color()[0] == 0 and light.get_color()[1] == 0 and light.get_color()[2] == 0]

    def activate(self, root: object = None) -> None:
        """
        Activates the light set on the root NodePath instance
        """

        if not root:
            root = runtime.render

        self.deactivate()
        for light in self.get_relevant_lights():
            self.set_transform(TransformStat.make_pos_hpr_scale(Vec3(0, 0, 0), Vec3(0, 90, 0), Vec3(1, 1, 1)))
            light_np = self.attach_new_node(light)
            root.set_light(light_np)
            self.__light_nps.append(light_np)

    def deactivate(self, root: object = None) -> None:
        """
        Deactivates the light set
        """

        if not root:
            root = runtime.render

        for light in self.__light_nps:
            if light:
                light.remove_node()

        root.clear_light()
        self.__light_nps = []

    def serialize_configuration(self) -> dict:
        """
        """

        data = Entity.serialize_configuration(self)
        data['ambientColor'] = Vec4(self.__ambient.get_color())

        for i in range(3):
            data['_directional_color%d' % (i + 1)] = Vec4(self.__directional_lights[i].get_color())
            data['_directional_hpr%d' % (i + 1)] = self.__directional_lights[i].get_direction()

        return data 

    def set_ambient_color(self, value: object) -> None:
        """
        """

        self.__ambient.set_color(value)

    def get_ambient_color(self) -> object:
        """
        """

        return self.__ambient.get_color()

    def get_directional(self, index: int) -> object:
        """
        Retrieves the directional at the requested index if it exists.
        Otherwise returns NoneType
        """

        if index > len(self.__directional_lights):
            return None

        return self.__directional_lights[index]

    def set_directional_hpr(self, index: int, value: object) -> None:
        """
        Sets the hpr of the directional at the requested index
        """

        directional = self.get_directional(index)
        if directional:
            directional.set_direction(value)

    def get_directional_hpr(self, index: int) -> object:
        """
        Retrieves the hpr of the directional at the requested index
        if it exists. Otherwise returns NoneType
        """

        directional = self.get_directional(index)
        if directional:
            return directional.get_direction()

        return None
                
    def set_directional_color(self, index: int, color: object) -> None:
        """
        Sets the color of the directional at the requested index
        """

        directional = self.get_directional(index)
        if directional:  
            directional.set_color(color)    

    def get_directional_color(self, index: int) -> object:
        """
        Retrieves the color of the directional at the requestsed index
        if it exists. Otherwise returns NoneType
        """

        directional = self.get_directional(index)
        if directional:
            return directional.get_color()

        return None


"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from panda3d.core import Shader, Filename

from panda3d_gemstone.logging.utilities import get_notify_category
from panda3d_gemstone.io.file_system import get_file_extension, path_exists

__shader_notify = get_notify_category('shaders')

class ShaderError(RuntimeError):
    """
    Base class for all shader errors
    """

class ShaderCompileError(ShaderError):
    """
    Fired when an error occures compiling a shader
    """

class MalformedShaderLoadRequest(ShaderError):
    """
    Fired when a malformed load_shader request is made
    """

def __kwargs_to_string(self, kwargs, sep=',') -> str:
    """
    Returns a list of keyword arguments as a string
    """

    output = ''
    for kwarg in kwargs:
        output += '%s: %s%s ' % (kwarg, kwargs[kwarg], sep)

    return output

__extension_map = {
    'glsl': Shader.SL_GLSL,
    'hlsl': Shader.SL_HLSL,
    'cg': Shader.SL_Cg,
    'spir': Shader.SL_SPIR_V
}

def is_supported_language(language: object) -> bool:
    """
    Returns true if the requested Panda3D idnetifier is supported
    """

    return language in __extension_map.values()

def shader_label_from_language(language: object) -> str:
    """
    Returns the shader language name based on the Panda3D identifier
    if found. Otherwise NoneType
    """

    found = None
    for shader_key in __extension_map:
        lang = __extension_map[shader_key]
        if lang == language:
            found = shader_key.upper()

    return found

def get_shader_language(shader_path: str) -> object:
    """
    Returns the Panda3D identifier for the shader file's language
    based on its file extension
    """

    ext = get_file_extension(shader_path)
    return __extension_map.get(ext, None)

def shader_file_supported(shader_path: str) -> bool:
    """
    Returns true if the shader file is supported by the Gemstone framework
    """

    return get_shader_language(shader_path) != None

def __matching_shader_languages(shader_paths: list) -> bool:
    """
    Returns true if all the shader paths provided match the same language
    """

    last_lang = None
    matching = True
    for shader_path in shader_paths:
        if shader_path is None:
            continue
            
        shader_lang = get_shader_language(shader_path)
        if last_lang != None and shader_lang != last_lang:
            matching = False
            break
        last_lang = shader_lang
    
    return matching

def __convert_path_list_filename_list(shader_paths: list) -> list:
    """
    Converts a list of string paths to a list of Panda3D Filename objects
    """

    converted = []
    for path in shader_paths:
        if path is None:
            continue
        
        converted.append(Filename(path))

    return converted

def make_shader(language: object, vertex: str, fragment: str, geometry: str = None, tess_control: str = None, tess_evaluation: str = None) -> object:
    """
    Creates a shader instance using string bodies instead of files
    """

    if not is_supported_language(language):
        raise ShaderCompileError('Failed to make shader. Language (%s) is not supported' % str(language))

    paths = [vertex, fragment, geometry, tess_control, tess_evaluation]
    paths = __convert_path_list_filename_list(paths)
    shader = Shader.make(language, *paths)

    # Verify no errors have occured
    if not shader or shader.get_error_flag():
        __shader_notify.warning('Failed to make shader. An error occured while compiling.')

    return shader

def make_compute_shader(language: object, vertex: str, fragment: str, geometry: str = None, tess_control: str = None, tess_evaluation: str = None) -> object:
    """
    Creates a compute shader instance using string bodies instead of files
    """

    if not is_supported_language(language):
        raise ShaderCompileError('Failed to make shader. Language (%s) is not supported' % str(language))

    paths = [vertex, fragment, geometry, tess_control, tess_evaluation]
    paths = __convert_path_list_filename_list(paths)
    shader = Shader.make_compute(language, *paths)

    # Verify no errors have occured
    if not shader or shader.get_error_flag():
        __shader_notify.warning('Failed to make compute shader. An error occured while compiling.')

    return shader

def load_shader(vertex: str, fragment: str, geometry: str = None, tess_control: str = None, tess_evaluation: str = None) -> object:
    """
    Loads a shader from the application's file system using the various shader program files.
    """

    # Verify the shader language is supported
    if not shader_file_supported(vertex):
        raise MalformedShaderLoadRequest('Invalid load request. Shader extension %s is not supported' % (get_file_extension(vertex)))

    # Verify all the requested shader files are the same language
    paths = [vertex, fragment, geometry, tess_control, tess_evaluation]
    if not __matching_shader_languages(paths):
        raise MalformedShaderLoadRequest('Invalid load request. Not all shader files are the same language. Shaders: [%s]' % (', '.join(paths)))

    shader_language = get_shader_language(vertex)
    paths = __convert_path_list_filename_list(paths)
    shader = Shader.load(shader_language, *paths)

    # Verify no errors have occured
    if not shader or shader.get_error_flag():
        __shader_notify.warning('Failed to load shader. An error occured while compiling.')

    return shader

def load_compute_shader(vertex: str, fragment: str, geometry: str = None, tess_control: str = None, tess_evaluation: str = None) -> object:
    """
    Loads a compute shader from the application's file system using the various shader program files.
    """

    # Verify the shader language is supported
    if not shader_file_supported(vertex):
        raise MalformedShaderLoadRequest('Invalid load compute request. Shader extension %s is not supported' % (get_file_extension(vertex)))

    # Verify all the requested shader files are the same language
    paths = [vertex, fragment, geometry, tess_control, tess_evaluation]
    if not __matching_shader_languages(paths):
        raise MalformedShaderLoadRequest('Invalid load compute request. Not all shader files are the same language. Shaders: [%s]' % (', '.join(paths)))

    shader_language = get_shader_language(vertex)
    paths = __convert_path_list_filename_list(paths)
    shader = Shader.load_compute(shader_language, *paths, **kwargs)

    # Verify no errors have occured
    if not shader or shader.get_error_flag():
        raise ShaderCompileError('Failed to load compute shader. An error occured while compiling.')

    return shader

def enable_shader_generator(nodepath: object) -> None:
    """
    Enables the default Panda3D shader generator on the requested node
    """

    assert hasattr(nodepath, 'set_shader_auto')
    nodepath.set_shader_auto()

def enable_global_shader_generator() -> None:
    """
    Globally enables the default Panda3D shader generator for the entire
    application 3d scene graph
    """

    from panda3d_gemstone.engine import runtime
    
    assert runtime.has_render()
    runtime.render.set_shader_auto()



"""
MIT License

Copyright (c) 2024 Jordan Maxwell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from PyQt5.QtGui import *

from panda3d_gemstone.framework.internal_object import InternalObject
from panda3d_gemstone.editor.level import EditorWorldManager, EditorScene

class QLevelTreeView(QTreeView):
    """
    """

    def __init__(self, *args, **kwargs):
        super(QTreeView, self).__init__(*args, **kwargs) 

    def build_tree(self, asset_tree: dict) -> None:
        """
        """

        root_model = QStandardItemModel()
        self.setModel(root_model)
        self._populate_tree(asset_tree, root_model.invisibleRootItem())
        root_model.setHeaderData(0, Qt.Horizontal, 'Level Browser')

        self.setDragDropMode(QAbstractItemView.DragOnly)
        self.setDragEnabled(True)
        self.setAcceptDrops(False)

    def _populate_tree(self, children: object, parent: object) -> None:
        """
        """

        if isinstance(children, str):
            parent.appendRow(QStandardItem(children))
            return

        for child in sorted(children):
            child_item = QStandardItem(child)
            parent.appendRow(child_item)
            if isinstance(children, dict):
                self._populate_tree(children[child], child_item)

class QLevelViewerWidget(QDockWidget, InternalObject):
    """
    """

    def __init__(self, *args, **kwargs):
        super(QDockWidget, self).__init__('Level Tree', *args, **kwargs)
        super(InternalObject, self).__init__()
        self._editor_scene = EditorScene.get_singleton()
        self._editor_world = EditorWorldManager.get_singleton()
        self.build_widget()

    def build_widget(self):
        """
        Builds the layer properties widget
        """

        # Build widget level tree
        level_tree = {
            'Scene': {},
            'Levels': {}
        }

        scene_objects = self._editor_scene.get_objects()
        for scene_object in scene_objects:
            name_parts = str(scene_object).split('/')
            object_name = name_parts[-1]

            tree_name = '%s (%s)' % (object_name, scene_object.__class__.__name__)
            level_tree['Scene'] = tree_name

        worlds = self._editor_world.worlds
        for world_key in worlds:
            world = worlds[world_key]
            if world_key not in level_tree['Levels']:
                level_tree['Levels'][world_key] = []
            
            for layer in world.get_layers():
                for entity in layer.get_entities():
                    tree_name = '%s (%s) (%s)' % (
                        entity.name, 
                        entity.__class__.__name__,
                        entity.get_layer().get_id())
                    
                    level_tree['Levels'][world_key].append(tree_name)
        
        self.tree = QLevelTreeView(self)
        self.tree.build_tree(level_tree)
        self.setWidget(self.tree)